@@ -1,18 +1,19 @@
-from binascii import hexlify
 from struct import pack, unpack
 import hashlib
 import time
 import sys
+import os
 import traceback
 
 import electrum
-from electrum.bitcoin import EncodeBase58Check, DecodeBase58Check, TYPE_ADDRESS, int_to_hex, var_int
+from electrum import bitcoin
+from electrum.bitcoin import TYPE_ADDRESS, int_to_hex, var_int
 from electrum.i18n import _
 from electrum.plugins import BasePlugin, hook
 from electrum.keystore import Hardware_KeyStore, parse_xpubkey
+from electrum.transaction import push_script, Transaction
 from ..hw_wallet import HW_PluginBase
-from electrum.util import format_satoshis_plain, print_error
-
+from electrum.util import format_satoshis_plain, print_error, is_verbose, bfh, bh2u
 
 try:
     import hid
@@ -20,11 +21,10 @@ try:
     from btchip.btchip import btchip
     from btchip.btchipUtils import compress_public_key,format_transaction, get_regular_input_script, get_p2sh_input_script
     from btchip.bitcoinTransaction import bitcoinTransaction
-    from btchip.btchipPersoWizard import StartBTChipPersoDialog
     from btchip.btchipFirmwareWizard import checkFirmware, updateFirmware
     from btchip.btchipException import BTChipException
     BTCHIP = True
-    BTCHIP_DEBUG = False
+    BTCHIP_DEBUG = is_verbose
 except ImportError:
     BTCHIP = False
 
@@ -51,7 +51,7 @@ class Ledger_Client():
     def i4b(self, x):
         return pack('>I', x)        
 
-    def get_xpub(self, bip32_path):
+    def get_xpub(self, bip32_path, xtype):
         self.checkDevice()
         # bip32_path is of the form 44'/0'/1'
         # S-L-O-W - we don't handle the fingerprint directly, so compute
@@ -59,42 +59,35 @@ class Ledger_Client():
         # This only happens once so it's bearable
         #self.get_client() # prompt for the PIN before displaying the dialog if necessary
         #self.handler.show_message("Computing master public key")
-        try:
-            splitPath = bip32_path.split('/')
-            if splitPath[0] == 'm':
-                splitPath = splitPath[1:]
-                bip32_path = bip32_path[2:]
-            fingerprint = 0
-            if len(splitPath) > 1:
-                prevPath = "/".join(splitPath[0:len(splitPath) - 1])
-                nodeData = self.dongleObject.getWalletPublicKey(prevPath)
-                publicKey = compress_public_key(nodeData['publicKey'])
-                h = hashlib.new('ripemd160')
-                h.update(hashlib.sha256(publicKey).digest())
-                fingerprint = unpack(">I", h.digest()[0:4])[0]
-            nodeData = self.dongleObject.getWalletPublicKey(bip32_path)
-            publicKey = compress_public_key(nodeData['publicKey'])
-            depth = len(splitPath)
-            lastChild = splitPath[len(splitPath) - 1].split('\'')
-            if len(lastChild) == 1:
-                childnum = int(lastChild[0])
-            else:
-                childnum = 0x80000000 | int(lastChild[0])
-            xpub = "0488B21E".decode('hex') + chr(depth) + self.i4b(fingerprint) + self.i4b(childnum) + str(nodeData['chainCode']) + str(publicKey)
-        except Exception, e:
-            #self.give_error(e, True)
-            return None
-        finally:
-            #self.handler.clear_dialog()
-            pass
-
-        return EncodeBase58Check(xpub)
+        if xtype in ['p2wpkh', 'p2wsh'] and not self.supports_native_segwit():
+            raise Exception("Firmware version too old for Segwit support. Please update at https://www.ledgerwallet.com")
+        if xtype in ['p2wpkh-p2sh', 'p2wsh-p2sh'] and not self.supports_segwit():
+            raise Exception("Firmware version too old for Segwit support. Please update at https://www.ledgerwallet.com")
+        splitPath = bip32_path.split('/')
+        if splitPath[0] == 'm':
+            splitPath = splitPath[1:]
+            bip32_path = bip32_path[2:]
+        fingerprint = 0
+        if len(splitPath) > 1:
+            prevPath = "/".join(splitPath[0:len(splitPath) - 1])
+            nodeData = self.dongleObject.getWalletPublicKey(prevPath)
+            publicKey = compress_public_key(nodeData['publicKey'])#
+            h = hashlib.new('ripemd160')
+            h.update(hashlib.sha256(publicKey).digest())
+            fingerprint = unpack(">I", h.digest()[0:4])[0]
+        nodeData = self.dongleObject.getWalletPublicKey(bip32_path)
+        publicKey = compress_public_key(nodeData['publicKey'])
+        depth = len(splitPath)
+        lastChild = splitPath[len(splitPath) - 1].split('\'')
+        childnum = int(lastChild[0]) if len(lastChild) == 1 else 0x80000000 | int(lastChild[0])
+        xpub = bitcoin.serialize_xpub(xtype, nodeData['chainCode'], publicKey, depth, self.i4b(fingerprint), self.i4b(childnum))
+        return xpub
 
     def has_detached_pin_support(self, client):
         try:
             client.getVerifyPinRemainingAttempts()
             return True
-        except BTChipException, e:
+        except BTChipException as e:
             if e.sw == 0x6d00:
                 return False
             raise e
@@ -103,32 +96,45 @@ class Ledger_Client():
         try:
             # Invalid SET OPERATION MODE to verify the PIN status
             client.dongle.exchange(bytearray([0xe0, 0x26, 0x00, 0x00, 0x01, 0xAB]))
-        except BTChipException, e:
+        except BTChipException as e:
             if (e.sw == 0x6982):
                 return False
             if (e.sw == 0x6A80):
                 return True
             raise e
 
+    def supports_multi_output(self):
+        return self.multiOutputSupported
+
+    def supports_segwit(self):
+        return self.segwitSupported
+
+    def supports_native_segwit(self):
+        return self.nativeSegwitSupported
+
     def perform_hw1_preflight(self):
         try:
-            firmware = self.dongleObject.getFirmwareVersion()['version'].split(".")
+            firmwareInfo = self.dongleObject.getFirmwareVersion()
+            firmware = firmwareInfo['version'].split(".")
+            self.multiOutputSupported = int(firmware[0]) >= 1 and int(firmware[1]) >= 1 and int(firmware[2]) >= 4
+            self.segwitSupported = (int(firmware[0]) >= 1 and int(firmware[1]) >= 1 and int(firmware[2]) >= 10) or (firmwareInfo['specialVersion'] == 0x20 and int(firmware[0]) == 1 and int(firmware[1]) == 0 and int(firmware[2]) >= 4)
+            self.nativeSegwitSupported = int(firmware[0]) >= 1 and int(firmware[1]) >= 1 and int(firmware[2]) >= 10
+
             if not checkFirmware(firmware):
-                self.dongleObject.close()
+                self.dongleObject.dongle.close()
                 raise Exception("HW1 firmware version too old. Please update at https://www.ledgerwallet.com")
             try:
                 self.dongleObject.getOperationMode()
-            except BTChipException, e:
+            except BTChipException as e:
                 if (e.sw == 0x6985):
-                    self.dongleObject.close()
-                    dialog = StartBTChipPersoDialog()
-                    dialog.exec_()
+                    self.dongleObject.dongle.close()
+                    self.handler.get_setup( )
                     # Acquire the new client on the next run
                 else:
                     raise e
-            if self.has_detached_pin_support(self.dongleObject) and not self.is_pin_validated(self.dongleObject) and (self.handler <> None):
+            if self.has_detached_pin_support(self.dongleObject) and not self.is_pin_validated(self.dongleObject) and (self.handler is not None):
                 remaining_attempts = self.dongleObject.getVerifyPinRemainingAttempts()
-                if remaining_attempts <> 1:
+                if remaining_attempts != 1:
                     msg = "Enter your Ledger PIN - remaining attempts : " + str(remaining_attempts)
                 else:
                     msg = "Enter your Ledger PIN - WARNING : LAST ATTEMPT. If the PIN is not correct, the dongle will be wiped."
@@ -137,7 +143,7 @@ class Ledger_Client():
                     raise Exception('Aborted by user - please unplug the dongle and plug it again before retrying')
                 pin = pin.encode()
                 self.dongleObject.verifyPin(pin)
-        except BTChipException, e:
+        except BTChipException as e:
             if (e.sw == 0x6faa):
                 raise Exception("Dongle is temporarily locked - please unplug it and replug it again")
             if ((e.sw & 0xFFF0) == 0x63c0):
@@ -172,11 +178,20 @@ class Ledger_KeyStore(Hardware_KeyStore):
         # device reconnects
         self.force_watching_only = False
         self.signing = False
+        self.cfg = d.get('cfg', {'mode':0,'pair':''})
+
+    def dump(self):
+        obj = Hardware_KeyStore.dump(self)
+        obj['cfg'] = self.cfg
+        return obj
 
     def get_derivation(self):
         return self.derivation        
 
     def get_client(self):
+        return self.plugin.get_client(self).dongleObject
+    
+    def get_client_electrum(self):
         return self.plugin.get_client(self)
     
     def give_error(self, message, clear_client = False):
@@ -201,6 +216,7 @@ class Ledger_KeyStore(Hardware_KeyStore):
 
     def sign_message(self, sequence, message, password):
         self.signing = True
+        message = message.encode('utf8')
         # prompt for the PIN before displaying the dialog if necessary
         client = self.get_client()
         address_path = self.get_derivation()[2:] + "/%d/%d"%sequence
@@ -209,26 +225,25 @@ class Ledger_KeyStore(Hardware_KeyStore):
             info = self.get_client().signMessagePrepare(address_path, message)
             pin = ""
             if info['confirmationNeeded']:
-                # TODO : handle different confirmation types. For the time being only supports keyboard 2FA
-                confirmed, p, pin = self.password_dialog()
-                if not confirmed:
-                    raise Exception('Aborted by user')
-                pin = pin.encode()
-                #self.plugin.get_client(self, True, True)
+                pin = self.handler.get_auth( info ) # does the authenticate dialog and returns pin
+                if not pin:
+                    raise UserWarning(_('Cancelled by user'))
+                pin = str(pin).encode()
             signature = self.get_client().signMessageSign(pin)
-        except BTChipException, e:
+        except BTChipException as e:
             if e.sw == 0x6a80:
                 self.give_error("Unfortunately, this message cannot be signed by the Ledger wallet. Only alphanumerical messages shorter than 140 characters are supported. Please remove any extra characters (tab, carriage return) and retry.")
             else:
                 self.give_error(e, True)
-        except Exception, e:
+        except UserWarning:
+            self.handler.show_error(_('Cancelled by user'))
+            return ''
+        except Exception as e:
             self.give_error(e, True)
         finally:
             self.handler.clear_dialog()
         self.signing = False
-
         # Parse the ASN.1 signature
-
         rLength = signature[3]
         r = signature[4 : 4 + rLength]
         sLength = signature[4 + rLength + 1]
@@ -237,11 +252,9 @@ class Ledger_KeyStore(Hardware_KeyStore):
             r = r[1:]
         if sLength == 33:
             s = s[1:]
-        r = str(r)
-        s = str(s)
-
         # And convert it
-        return chr(27 + 4 + (signature[0] & 0x01)) + r + s
+        return bytes([27 + 4 + (signature[0] & 0x01)]) + r + s
+
 
     def sign_transaction(self, tx, password):
         if tx.is_complete():
@@ -260,19 +273,31 @@ class Ledger_KeyStore(Hardware_KeyStore):
         output = None
         outputAmount = None
         p2shTransaction = False
+        segwitTransaction = False
         pin = ""
         self.get_client() # prompt for the PIN before displaying the dialog if necessary
 
         # Fetch inputs of the transaction to sign
         derivations = self.get_tx_derivations(tx)
         for txin in tx.inputs():
-            if txin.get('is_coinbase'):
+            if txin['type'] == 'coinbase':
                 self.give_error("Coinbase not supported")     # should never happen
 
-            if len(txin['pubkeys']) > 1:
+            if txin['type'] in ['p2sh']:
                 p2shTransaction = True
 
-            for i, x_pubkey in enumerate(txin['x_pubkeys']):
+            if txin['type'] in ['p2wpkh-p2sh', 'p2wsh-p2sh']:
+                if not self.get_client_electrum().supports_segwit():
+                    self.give_error("Firmware version too old to support segwit. Please update at https://www.ledgerwallet.com")
+                segwitTransaction = True
+
+            if txin['type'] in ['p2wpkh', 'p2wsh']:
+                if not self.get_client_electrum().supports_native_segwit():
+                    self.give_error("Firmware version too old to support native segwit. Please update at https://www.ledgerwallet.com")
+                segwitTransaction = True
+
+            pubkeys, x_pubkeys = tx.get_sorted_pubkeys(txin)
+            for i, x_pubkey in enumerate(x_pubkeys):
                 if x_pubkey in derivations:
                     signingPos = i
                     s = derivations.get(x_pubkey)
@@ -281,14 +306,15 @@ class Ledger_KeyStore(Hardware_KeyStore):
             else:
                 self.give_error("No matching x_key for sign_transaction") # should never happen
 
-            inputs.append([txin['prev_tx'].raw, txin['prevout_n'], txin.get('redeemScript'), txin['prevout_hash'], signingPos ])
+            redeemScript = Transaction.get_preimage_script(txin)
+            inputs.append([txin['prev_tx'].raw, txin['prevout_n'], redeemScript, txin['prevout_hash'], signingPos, txin.get('sequence', 0xffffffff - 1) ])
             inputsPaths.append(hwAddress)
-            pubKeys.append(txin['pubkeys'])
+            pubKeys.append(pubkeys)
 
         # Sanity check
         if p2shTransaction:
-            for txinput in tx.inputs():
-                if len(txinput['pubkeys']) < 2:
+            for txin in tx.inputs():
+                if txin['type'] != 'p2sh':
                     self.give_error("P2SH / regular input mixed in same transaction not supported") # should never happen
 
         txOutput = var_int(len(tx.outputs()))
@@ -296,18 +322,19 @@ class Ledger_KeyStore(Hardware_KeyStore):
             output_type, addr, amount = txout
             txOutput += int_to_hex(amount, 8)
             script = tx.pay_script(output_type, addr)
-            txOutput += var_int(len(script)/2)
+            txOutput += var_int(len(script)//2)
             txOutput += script
-        txOutput = txOutput.decode('hex')
+        txOutput = bfh(txOutput)
 
         # Recognize outputs - only one output and one change is authorized
         if not p2shTransaction:
-            if len(tx.outputs()) > 2: # should never happen
-                self.give_error("Transaction with more than 2 outputs not supported")
+            if not self.get_client_electrum().supports_multi_output():
+                if len(tx.outputs()) > 2:
+                    self.give_error("Transaction with more than 2 outputs not supported")
             for _type, address, amount in tx.outputs():
                 assert _type == TYPE_ADDRESS
                 info = tx.output_info.get(address)
-                if info is not None:
+                if (info is not None) and (len(tx.outputs()) != 1):
                     index, xpubs, m = info
                     changePath = self.get_derivation()[2:] + "/%d/%d"%index
                     changeAmount = amount
@@ -319,110 +346,93 @@ class Ledger_KeyStore(Hardware_KeyStore):
         try:
             # Get trusted inputs from the original transactions
             for utxo in inputs:                
-                if not p2shTransaction:
-                    txtmp = bitcoinTransaction(bytearray(utxo[0].decode('hex')))             
-                    chipInputs.append(self.get_client().getTrustedInput(txtmp, utxo[1]))
-                    redeemScripts.append(txtmp.outputs[utxo[1]].script)                    
+                sequence = int_to_hex(utxo[5], 4)
+                if segwitTransaction:
+                    txtmp = bitcoinTransaction(bfh(utxo[0]))                    
+                    tmp = bfh(utxo[3])[::-1]
+                    tmp += bfh(int_to_hex(utxo[1], 4))
+                    tmp += txtmp.outputs[utxo[1]].amount
+                    chipInputs.append({'value' : tmp, 'witness' : True, 'sequence' : sequence})
+                    redeemScripts.append(bfh(utxo[2]))
+                # FIXME: getTrustedInput fails with native segwit transactions (firmware issue)
+                #elif not p2shTransaction:
+                #    txtmp = bitcoinTransaction(bfh(utxo[0]))
+                #    trustedInput = self.get_client().getTrustedInput(txtmp, utxo[1])
+                #    trustedInput['sequence'] = sequence
+                #    chipInputs.append(trustedInput)
+                #    redeemScripts.append(txtmp.outputs[utxo[1]].script)
                 else:
-                    tmp = utxo[3].decode('hex')[::-1].encode('hex')
-                    tmp += int_to_hex(utxo[1], 4)
-                    chipInputs.append({'value' : tmp.decode('hex')})
-                    redeemScripts.append(bytearray(utxo[2].decode('hex')))
+                    tmp = bfh(utxo[3])[::-1]
+                    tmp += bfh(int_to_hex(utxo[1], 4))
+                    chipInputs.append({'value' : tmp, 'sequence' : sequence})
+                    redeemScripts.append(bfh(utxo[2]))
 
             # Sign all inputs
             firstTransaction = True
             inputIndex = 0
             rawTx = tx.serialize()
-            while inputIndex < len(inputs):
-                self.get_client().startUntrustedTransaction(firstTransaction, inputIndex,
+            self.get_client().enableAlternate2fa(False)
+            if segwitTransaction:
+                self.get_client().startUntrustedTransaction(True, inputIndex,
                                                             chipInputs, redeemScripts[inputIndex])
-                if not p2shTransaction:
-                    outputData = self.get_client().finalizeInput(output, format_satoshis_plain(outputAmount),
-                        format_satoshis_plain(tx.get_fee()), changePath, bytearray(rawTx.decode('hex')))
-                else:
-                    outputData = self.get_client().finalizeInputFull(txOutput)
-                    outputData['outputData'] = txOutput
-
-                if firstTransaction:
-                    transactionOutput = outputData['outputData']
+                outputData = self.get_client().finalizeInputFull(txOutput)
+                outputData['outputData'] = txOutput
+                transactionOutput = outputData['outputData']
                 if outputData['confirmationNeeded']:
-                    # TODO : handle different confirmation types. For the time being only supports keyboard 2FA
+                    outputData['address'] = output
                     self.handler.clear_dialog()
-                    if 'keycardData' in outputData:
-                        pin2 = ""
-                        for keycardIndex in range(len(outputData['keycardData'])):
-                            msg = "Do not enter your device PIN here !\r\n\r\n" + \
-                                "Your Ledger Wallet wants to talk to you and tell you a unique second factor code.\r\n" + \
-                                "For this to work, please match the character between stars of the output address using your security card\r\n\r\n" + \
-                                "Output address : "
-                            for index in range(len(output)):
-                                if index == outputData['keycardData'][keycardIndex]:
-                                    msg = msg + "*" + output[index] + "*"
-                                else:
-                                    msg = msg + output[index]
-                            msg = msg + "\r\n"
-                            confirmed, p, pin = self.password_dialog(msg)
-                            if not confirmed:
-                                raise Exception('Aborted by user')
-                            try:
-                                pin2 = pin2 + chr(int(pin[0], 16))
-                            except:
-                                raise Exception('Invalid PIN character')
-                        pin = pin2
-                    else:
-                        confirmed, p, pin = self.password_dialog()
-                        if not confirmed:
-                            raise Exception('Aborted by user')
-                        pin = pin.encode()
-                        #self.plugin.get_client(self, True, True)
-                    self.handler.show_message("Signing ...")
-                else:
-                    # Sign input with the provided PIN
-                    inputSignature = self.get_client().untrustedHashSign(inputsPaths[inputIndex],
-                    pin)
+                    pin = self.handler.get_auth( outputData ) # does the authenticate dialog and returns pin
+                    if not pin:
+                        raise UserWarning()
+                    if pin != 'paired':
+                        self.handler.show_message(_("Confirmed. Signing Transaction..."))
+                while inputIndex < len(inputs):                
+                    singleInput = [ chipInputs[inputIndex] ]
+                    self.get_client().startUntrustedTransaction(False, 0,
+                                                            singleInput, redeemScripts[inputIndex])
+                    inputSignature = self.get_client().untrustedHashSign(inputsPaths[inputIndex], pin)
                     inputSignature[0] = 0x30 # force for 1.4.9+
                     signatures.append(inputSignature)
                     inputIndex = inputIndex + 1
-                firstTransaction = False
+            else:
+                while inputIndex < len(inputs):
+                    self.get_client().startUntrustedTransaction(firstTransaction, inputIndex,
+                                                            chipInputs, redeemScripts[inputIndex])
+                    outputData = self.get_client().finalizeInputFull(txOutput)
+                    outputData['outputData'] = txOutput
+                    if firstTransaction:
+                        transactionOutput = outputData['outputData']
+                    if outputData['confirmationNeeded']:
+                        outputData['address'] = output
+                        self.handler.clear_dialog()
+                        pin = self.handler.get_auth( outputData ) # does the authenticate dialog and returns pin
+                        if not pin:
+                            raise UserWarning()
+                        if pin != 'paired':
+                            self.handler.show_message(_("Confirmed. Signing Transaction..."))
+                    else:
+                        # Sign input with the provided PIN
+                        inputSignature = self.get_client().untrustedHashSign(inputsPaths[inputIndex], pin, lockTime=tx.locktime)
+                        inputSignature[0] = 0x30 # force for 1.4.9+
+                        signatures.append(inputSignature)
+                        inputIndex = inputIndex + 1
+                    if pin != 'paired':
+                        firstTransaction = False
+        except UserWarning:
+            self.handler.show_error(_('Cancelled by user'))
+            return
         except BaseException as e:
             traceback.print_exc(file=sys.stdout)
             self.give_error(e, True)
         finally:
             self.handler.clear_dialog()
 
-        # Reformat transaction
-        inputIndex = 0
-        while inputIndex < len(inputs):
-            if p2shTransaction:
-                signaturesPack = [signatures[inputIndex]] * len(pubKeys[inputIndex])
-                inputScript = get_p2sh_input_script(redeemScripts[inputIndex], signaturesPack)
-                preparedTrustedInputs.append([ ("\x00" * 4) + chipInputs[inputIndex]['value'], inputScript ])                
-            else:
-                inputScript = get_regular_input_script(signatures[inputIndex], pubKeys[inputIndex][0].decode('hex'))
-                preparedTrustedInputs.append([ chipInputs[inputIndex]['value'], inputScript ])
-            inputIndex = inputIndex + 1
-        updatedTransaction = format_transaction(transactionOutput, preparedTrustedInputs)
-        updatedTransaction = hexlify(updatedTransaction)
-        tx.update_signatures(updatedTransaction)
+        for i, txin in enumerate(tx.inputs()):
+            signingPos = inputs[i][4]
+            txin['signatures'][signingPos] = bh2u(signatures[i])
+        tx.raw = tx.serialize()
         self.signing = False
 
-    def password_dialog(self, msg=None):
-        if not msg:
-            msg = _("Do not enter your device PIN here !\r\n\r\n" \
-                    "Your Ledger Wallet wants to talk to you and tell you a unique second factor code.\r\n" \
-                    "For this to work, please open a text editor " \
-                    "(on a different computer / device if you believe this computer is compromised) " \
-                    "and put your cursor into it, unplug your Ledger Wallet and plug it back in.\r\n" \
-                    "It should show itself to your computer as a keyboard " \
-                    "and output the second factor along with a summary of " \
-                    "the transaction being signed into the text-editor.\r\n\r\n" \
-                    "Check that summary and then enter the second factor code here.\r\n" \
-                    "Before clicking OK, re-plug the device once more (unplug it and plug it again if you read the second factor code on the same computer)")
-        response = self.handler.get_word(msg)
-        if response is None:
-            return False, None, None
-        return True, response, response
-
 
 class LedgerPlugin(HW_PluginBase):
     libraries_available = BTCHIP
@@ -438,6 +448,7 @@ class LedgerPlugin(HW_PluginBase):
                  ]
 
     def __init__(self, parent, config, name):
+        self.segwit = config.get("segwit")
         HW_PluginBase.__init__(self, parent, config, name)
         if self.libraries_available:
             self.device_manager().register_devices(self.DEVICE_IDS)
@@ -462,7 +473,7 @@ class LedgerPlugin(HW_PluginBase):
         self.handler = handler
 
         client = self.get_btchip_device(device)
-        if client <> None:
+        if client is not None:
             client = Ledger_Client(client)
         return client
 
@@ -470,18 +481,15 @@ class LedgerPlugin(HW_PluginBase):
         devmgr = self.device_manager()
         device_id = device_info.device.id_
         client = devmgr.client_by_id(device_id)
-        #client.handler = wizard
         client.handler = self.create_handler(wizard)
-        #client.get_xpub('m')
-        client.get_xpub("m/44'/0'") # TODO replace by direct derivation once Nano S > 1.1
+        client.get_xpub("m/44'/0'", 'standard') # TODO replace by direct derivation once Nano S > 1.1
 
-    def get_xpub(self, device_id, derivation, wizard):
+    def get_xpub(self, device_id, derivation, xtype, wizard):
         devmgr = self.device_manager()
         client = devmgr.client_by_id(device_id)
-        #client.handler = wizard
         client.handler = self.create_handler(wizard)
         client.checkDevice()
-        xpub = client.get_xpub(derivation)
+        xpub = client.get_xpub(derivation, xtype)
         return xpub
 
     def get_client(self, keystore, force_pair=True):
@@ -489,13 +497,11 @@ class LedgerPlugin(HW_PluginBase):
         #assert self.main_thread != threading.current_thread()
         devmgr = self.device_manager()
         handler = keystore.handler
-        handler = keystore.handler
         with devmgr.hid_lock:
             client = devmgr.client_for_keystore(self, handler, keystore, force_pair)        
         # returns the client for a given keystore. can use xpub
         #if client:
         #    client.used()
-        if client <> None:
-            client.checkDevice()
-            client = client.dongleObject            
+        if client is not None:
+            client.checkDevice()                    
         return client
